import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.feature_selection import RFE
from sklearn.model_selection import cross_val_score
from kowalsky.logs.utils import LivePyPlot
from mlxtend.feature_selection import SequentialFeatureSelector as SFS
from mlxtend.plotting import plot_sequential_feature_selection as plot_sfs
import matplotlib.pyplot as plt
from sklearn.metrics import get_scorer
from sklearn.base import clone
from .logs.loggers import get_logger
import numpy as np
import math


def calc_score(model, X, y, type='cv', scoring=None, cv=5):
    if type == 'cv':
        return abs(cross_val_score(model, X, y, scoring=scoring, cv=cv, n_jobs=-1, verbose=10).mean())
    elif type == 'test':
        X_train, X_test, y_train, y_test = train_test_split(X, y)
        model.fit(X_train, y_train)
        preds = model.predict(X_test)
        return abs(get_scorer(scoring)._score_func(y_test, preds))
    else:
        raise Exception(f'Wrong type: {type} (use cv/test)')


class XRFE(RFE):

    def __init__(self, estimator, n_features_to_select=None, step=1, logging=None, logging_params={}):
        self.estimator = estimator
        self.n_features_to_select = n_features_to_select
        self.step = step
        self.logger = get_logger(logging, 'XRFE', logging_params)

    def fit(self, X, y):
        n_features = X.shape[1]
        all_features = X.columns
        support_ = np.ones(n_features, dtype=bool)
        ranking_ = np.ones(n_features, dtype=int)

        while np.sum(support_) > self.n_features_to_select:
            features = np.arange(n_features)[support_]
            estimator = clone(self.estimator)

            self.logger("Training %d cols" % np.sum(support_))
            estimator.fit(X.iloc[:, features], y)
            importances = estimator.feature_importances_
            ranks = np.argsort(importances)
            ranks = np.ravel(ranks)

            # Eliminate the worse features
            threshold = min(self.step, np.sum(support_) - self.n_features_to_select)

            support_[features[ranks][:threshold]] = False
            ranking_[np.logical_not(support_)] += 1

        features = np.arange(n_features)[support_]
        self.estimator_ = clone(self.estimator)
        self.estimator_.fit(X.loc[:, all_features[features]], y)
        self.n_features_ = support_.sum()
        self.support_ = support_
        self.ranking_ = ranking_

        self.logger('Support: ', support_)
        self.logger('Ranking: ', ranking_)
        return self


class XRFECV(RFE):

    def __init__(self, estimator, n_features_to_select=None, step=1, scorer=None, logging=None, logging_params={}):
        self.estimator = estimator
        self.n_features_to_select = n_features_to_select
        self.step = step
        self.scorer = scorer
        self.logger = get_logger(logging, 'XRFECV', logging_params)

    def fit(self, X, y):
        n_features = X.shape[1]
        all_features = X.columns
        support_ = np.ones(n_features, dtype=bool)
        ranking_ = np.ones(n_features, dtype=int)
        self.scores = []

        while np.sum(support_) > self.n_features_to_select:
            features = np.arange(n_features)[support_]
            estimator = clone(self.estimator)

            self.logger("Training %d cols" % np.sum(support_))
            estimator.fit(X.iloc[:, features], y)
            importances = estimator.feature_importances_
            ranks = np.argsort(importances)
            ranks = np.ravel(ranks)

            # Eliminate the worse features
            threshold = min(self.step, np.sum(support_) - self.n_features_to_select)

            support_[features[ranks][:threshold]] = False
            ranking_[np.logical_not(support_)] += 1

            if self.scorer is not None:
                res = cross_val_score(estimator, X.iloc[:, features], y, n_jobs=-1)
                self.scores.append(res)
                self.logger(f"Scores: {np.mean(res)}")

        features = np.arange(n_features)[support_]
        self.estimator_ = clone(self.estimator)
        self.estimator_.fit(X.loc[:, all_features[features]], y)
        self.n_features_ = support_.sum()
        self.support_ = support_
        self.ranking_ = ranking_

        self.logger('Support: ', support_)
        self.logger('Ranking: ', ranking_)
        return self


def rfe_analysis(model, y_column, scoring, direction, cv=3, ranking=None, path=None, ds=None,
                 sample_size=None, n_features_to_select=5, precision=1e-3, stratify=False, eval_type='cv',
                 logging=None, logging_params={}):

    logger = get_logger(logging, 'feature_analysis', logging_params)

    if ds is None:
        ds = pd.read_csv(path)

    if sample_size is not None:
        ds, _ = train_test_split(ds, train_size=sample_size, stratify=ds[y_column] if stratify else None)

    X, y = ds.drop(y_column, axis=1), ds[y_column]

    if ranking is None:
        rfe = RFE(model, step=1, verbose=10, n_features_to_select=n_features_to_select)
        try:
            rfe.fit(X, y)
        except KeyboardInterrupt:
            logger("Stopped with keyboard")
            return
        ranking = rfe.ranking_
    else:
        ranking = np.array(ranking)

    logger(pd.DataFrame({
        'column': X.columns,
        'ranking': ranking
    }).sort_values(by='ranking').values)

    live = LivePyPlot(direction=direction, show_true_value=True)

    best_score = -math.inf if direction == 'maximize' else math.inf
    best_columns = None

    try:
        for rank in range(1, np.max(ranking) + 1):
            rank_columns = X.columns[ranking <= rank]
            score = calc_score(model, X[rank_columns], y, type=eval_type, cv=cv, scoring=scoring)
            live(score, sum(ranking <= rank))
            logger(score, sum(ranking <= rank))

            is_better_score = score > best_score + precision if direction == 'maximize' else score < best_score - precision
            if is_better_score:
                best_score = score
                best_columns = ranking <= rank
    except KeyboardInterrupt:
        logger("Stopped with keyboard")

    live.clear()

    logger('Best score', best_score)
    logger('Best columns', best_columns)
    logger('Ranking', ranking)
    return best_columns, ranking


def sfs_analysis(model, y_label, scoring, k_features, cv=3, forward=True, floating=False, path=None,
                 sample_size=None, stratify=False, ds=None, logging=None, logging_params={}):

    logger = get_logger(logging, 'sfs_feature_analysis', logging_params)

    if ds is None:
        ds = pd.read_csv(path)

    if sample_size is not None:
        ds, _ = train_test_split(ds, train_size=sample_size, stratify=ds[y_label] if stratify else None)

    X, y = ds.drop(y_label, axis=1), ds[y_label]

    selector = SFS(model, forward=forward, floating=floating, verbose=10,
                   scoring=scoring, cv=cv, n_jobs=-1, k_features=k_features)
    try:
        selector.fit(X, y)
    except KeyboardInterrupt:
        print("Stopped with keyboard")

    plot_sfs(selector.get_metric_dict(), kind='std_dev')

    plt.grid()
    plt.show()

    logger(selector.get_metric_dict())

    return selector.get_metric_dict()



import pandas as pd
import numpy as np
from sklearn.preprocessing import StandardScaler, MinMaxScaler
import missingno as msno
from kowalsky.logs.utils import calc_percent
from sklearn.model_selection import train_test_split


def corr(ds, y_col):
    return abs(ds.corr()[y_col]).sort_values()


def handle_outliers(df_raw, columns, drop=True, upper_quantile=.95,
                    lower_quantile=.05):
    df = df_raw.copy()

    out_cols = []
    for ix, col_value in enumerate(columns):
        col_in_df = isinstance(col_value, str)

        if col_in_df:
            if col_value not in df: continue
            column = df[col_value]
        else:
            column = col_value

        upper_lim = column.quantile(upper_quantile)
        lower_lim = column.quantile(lower_quantile)

        if not drop:
            if col_in_df:
                df.loc[(column > upper_lim), col_value] = upper_lim
                df.loc[(column < lower_lim), col_value] = lower_lim
            else:
                out_col = column.copy()
                out_col[column > upper_lim] = upper_lim
                out_col[column < lower_lim] = lower_lim
                out_cols.append(out_col)
        else:
            ix = (column < upper_lim) & (column > lower_lim)
            df = df[ix]
            if not col_in_df:
                out_col = column.copy()
                out_cols.append(out_col[ix])

    return df, out_cols


def transform(df_raw, columns, fn=np.log1p):
    df = df_raw.copy()
    for ix, column in enumerate(columns):
        if column in df:
            df[column] = df[column].transform(fn)

    return df


def drop(df_raw, columns):
    df = df_raw.copy()
    for column in columns:
        if column in df:
            df.drop(column, axis=1, inplace=True)

    return df


def group_by_mean(df_raw, pairs):
    df = df_raw.copy()
    for group_col, agr_col in pairs:
        if group_col in df and agr_col in df:
            df = pd.merge(df, df.groupby(group_col)[agr_col].mean(),
                          left_on=group_col, right_on=group_col, suffixes=('', f'_{group_col}_mean'))

    return df


def group_by_max(df_raw, pairs):
    df = df_raw.copy()
    for group_col, agr_col in pairs:
        if group_col in df and agr_col in df:
            df = pd.merge(df, df.groupby(group_col)[agr_col].max(),
                          left_on=group_col, right_on=group_col, suffixes=('', f'_{group_col}_max'))
    return df


def group_by_min(df_raw, pairs):
    df = df_raw.copy()
    for group_col, agr_col in pairs:
        if group_col in df and agr_col in df:
            df = pd.merge(df, df.groupby(group_col)[agr_col].min(),
                          left_on=group_col, right_on=group_col, suffixes=('', f'_{group_col}_min'))
    return df


def scale(df_raw, columns, minMax=False):
    df = df_raw.copy()

    if minMax:
        scaler = MinMaxScaler()
    else:
        scaler = StandardScaler()

    for col in columns:
        if col in df_raw:
            df[col] = scaler.fit_transform(np.array(df[col]).reshape(-1, 1)).reshape(-1)
    return df


def describe_missing_values(df):
    msno.bar(df)
    msno.matrix(df)
    msno.heatmap(df)


def read_dataset(ds, path, y_column, feature_selection_support, feature_selection_cols,
                 ignore=[], sample_size=None, stratify=True):
    if ds is None:
        ds = pd.read_csv(path)

    if sample_size is not None:
        ds, _ = train_test_split(ds, train_size=sample_size, stratify=ds[y_column] if stratify else None)

    X, y = ds.drop([y_column] + ignore, axis=1), ds[y_column]

    if feature_selection_support is None and feature_selection_cols is None:
        feature_selection_cols = X.columns

    return X[X.columns[feature_selection_support] if feature_selection_support is not None else feature_selection_cols], y





import math
import numpy as np
import matplotlib.pyplot as plt
from tensorflow import keras

class OneCycleScheduler(keras.callbacks.Callback):
    def __init__(self, iterations, max_rate, start_rate=None,
                 last_iterations=None, last_rate=None):
        self.iterations = iterations
        self.max_rate = max_rate
        self.start_rate = start_rate or max_rate / 10
        self.last_iterations = last_iterations or iterations // 10 + 1
        self.half_iteration = (iterations - self.last_iterations) // 2
        self.last_rate = last_rate or self.start_rate / 1000
        self.iteration = 0
    def _interpolate(self, iter1, iter2, rate1, rate2):
        return ((rate2 - rate1) * (self.iteration - iter1)
                / (iter2 - iter1) + rate1)
    def on_batch_begin(self, batch, logs):
        if self.iteration < self.half_iteration:
            rate = self._interpolate(0, self.half_iteration, self.start_rate, self.max_rate)
        elif self.iteration < 2 * self.half_iteration:
            rate = self._interpolate(self.half_iteration, 2 * self.half_iteration,
                                     self.max_rate, self.start_rate)
        else:
            rate = self._interpolate(2 * self.half_iteration, self.iterations,
                                     self.start_rate, self.last_rate)
        self.iteration += 1
        # K.set_value(self.model.optimizer.lr, rate)

class ExponentialLearningRate(keras.callbacks.Callback):
    def __init__(self, factor):
        self.factor = factor
        self.rates = []
        self.losses = []
    def on_batch_end(self, batch, logs):
        # self.rates.append(K.get_value(self.model.optimizer.lr))
        self.losses.append(logs["loss"])
        # K.set_value(self.model.optimizer.lr, self.model.optimizer.lr * self.factor)

def find_learning_rate(backend, model, X, y, epochs=1, batch_size=32, min_rate=10**-5, max_rate=10):
    init_weights = model.get_weights()
    iterations = math.ceil(len(X) / batch_size) * epochs
    factor = np.exp(np.log(max_rate / min_rate) / iterations)
    init_lr = backend.get_value(model.optimizer.lr)
    backend.set_value(model.optimizer.lr, min_rate)
    exp_lr = ExponentialLearningRate(factor)
    history = model.fit(X, y, epochs=epochs, batch_size=batch_size,
                        callbacks=[exp_lr])
    backend.set_value(model.optimizer.lr, init_lr)
    model.set_weights(init_weights)
    return exp_lr.rates, exp_lr.losses

def plot_lr_vs_loss(rates, losses):
    plt.plot(rates, losses)
    plt.gca().set_xscale('log')
    plt.hlines(min(losses), min(rates), max(rates))
    plt.axis([min(rates), max(rates), min(losses), (losses[0] + min(losses)) / 2])
    plt.xlabel("Learning rate")
    plt.ylabel("Loss")


import optuna
from lightgbm import LGBMRegressor
from lightgbm import LGBMClassifier
from xgboost import XGBRegressor
from xgboost import XGBClassifier
from optuna.samplers import TPESampler
from sklearn.ensemble import RandomForestClassifier
from sklearn.ensemble import RandomForestRegressor
from sklearn.model_selection import train_test_split
from sklearn.tree import DecisionTreeClassifier
from sklearn.tree import DecisionTreeRegressor
from sklearn.tree import ExtraTreeRegressor
from sklearn.tree import ExtraTreeClassifier
from sklearn.ensemble import BaggingClassifier
from sklearn.ensemble import BaggingRegressor
from sklearn.ensemble import AdaBoostClassifier
from sklearn.ensemble import AdaBoostRegressor
from sklearn.neighbors import KNeighborsClassifier
from sklearn.neighbors import KNeighborsRegressor
from sklearn.svm import SVR
from sklearn.svm import SVC
from catboost import CatBoostClassifier
from catboost import CatBoostRegressor
from sklearn.metrics import get_scorer as get_sklearn_scorer
from kowalsky.logs.utils import LivePyPlot
from .logs.loggers import get_logger
from .model import EarlyStopping
from .model import EarlyStoppingError
import numpy as np
from .df import read_dataset
from .metrics import neg_rmse
from sklearn.metrics import get_scorer

family_params = {
    'hgb': {
        'learning_rate': ('uniform', 0.0001, 2),
        'max_iter': ('uniform', 50, 3000),
        'max_depth': ('int', 1, 25),
        'max_bins': ('int', 10, 400),
        'max_leaf_nodes': ('int', 10, 60),
        'min_samples_leaf': ('uniform', 0.00001, 0.5),
        'l2_regularization': ('uniform', 0., 20.)
    },
    'gb': {
        'learning_rate': ('uniform', 0.0001, 2),
        'n_estimators': ('int', 1, 600),
        'max_depth': ('int', 1, 25),
        'min_samples_split': ('uniform', 0.00001, 0.5),
        'min_samples_leaf': ('uniform', 0.00001, 0.5),
        'max_features': ('categorical', ['log2', 'sqrt']),
        'criterion': ('categorical', ['friedman_mse', 'mae']),
        'subsample': ('uniform', 0.3, 1.0)
    },
    'lgb': {
        'learning_rate': ('uniform', 0.0001, 1),
        'n_estimators': ('int', 1, 6200),
        'max_depth': ('int', 1, 25),
        'num_leaves': ('int', 2, 3000),
        'min_child_samples': ('int', 3, 200),
        'max_bin': ('int', 10, 255),
        'lambda_l1': ('uniform', 0, 10),
        'lambda_l2': ('uniform', 0, 10),
        'max_cat_threshold': ('int', 1, 60),
        'min_gain_to_split': ('uniform', 0, 10),
        'min_data_in_leaf': ('int', 5, 60)
    },
    'dt': {
        'max_depth': ('int', 2, 25),
        'min_samples_split': ('int', 2, 20),
        'min_weight_fraction_leaf': ('uniform', 0.0, 0.5),
        'min_samples_leaf': ('int', 1, 15)
    },
    'xgb': {
        'learning_rate': ('uniform', 0.0000001, 1),
        'n_estimators': ('int', 2, 1000),
        'max_depth': ('int', 2, 25),
        'gamma': ('uniform', 0.0000001, 20),
        'alpha': ('uniform', 0, 20),
        'colsample_bytree': ('uniform', 0.3, 1),
        'lambda': ('uniform', 0, 10),
        'max_delta_step': ('uniform', 0, 20),
        'min_child_weight': ('uniform', 0, 20),
        'subsample': ('uniform', 0.2, 1)
    },
    'rf': {
        'min_samples_leaf': ('int', 1, 15),
        'min_samples_split': ('uniform', 0.05, 1.0),
        'n_estimators': ('int', 2, 800),
        'max_depth': ('int', 2, 25),
        'random_state': 666
    },
    'et': {
        'min_samples_leaf': ('int', 1, 15),
        'min_samples_split': ('uniform', 0.05, 1.0),
        'max_depth': ('int', 2, 25),
        'random_state': 666
    },
    'bagg': {
        'n_estimators': ('int', 2, 300),
        'max_samples': ('int', 1, 400),
        'random_state': 666
    },
    'kn': {
        'n_neighbors': ('int', 2, 100)
    },
    'ada': {
        'n_estimators': ('int', 2, 800),
        'learning_rate': ('uniform', 0.0001, 1.0)
    },
    'svm': {
        'kernel': ('categorical', ['linear', 'poly']),
        'tol': ('uniform', 1e-5, 1),
        'C': ('loguniform', 1e-10, 1e10)
    },
    'cb': {
        'learning_rate': ('uniform', 0.0001, 1.0),
        'depth': ('int', 1, 16),
        'iterations': ('int', 100, 2000),
        'l2_leaf_reg': ('uniform', 0, 10),
        'border_count': ('int', 1, 148),
        'subsample': ('uniform', 0.3, 1),
    }
}

models = {

    # Gradient Boosts
    'xgbR': (XGBRegressor, 'xgb'),
    'xgbC': (XGBClassifier, 'xgb'),
    'lgbR': (LGBMRegressor, 'lgb'),
    'lgbC': (LGBMClassifier, 'lgb'),

    # Trees
    'rfR': (RandomForestRegressor, 'rf'),
    'rfC': (RandomForestClassifier, 'rf'),
    'dtR': (DecisionTreeRegressor, 'dt'),
    'dtC': (DecisionTreeClassifier, 'dt'),
    'etR': (ExtraTreeRegressor, 'et'),
    'etC': (ExtraTreeClassifier, 'et'),

    # Ensemble
    'baggC': (BaggingClassifier, 'bagg'),
    'baggR': (BaggingRegressor, 'bagg'),
    'adaR': (AdaBoostRegressor, 'ada'),
    'adaC': (AdaBoostClassifier, 'ada'),
    'cbR': (CatBoostRegressor, 'cb'),
    'cbC': (CatBoostClassifier, 'cb'),

    # KNeighbors
    'knC': (KNeighborsClassifier, 'kn'),
    'knR': (KNeighborsRegressor, 'kn'),

    # SVM
    'svR': (SVR, 'svm'),
    'svC': (SVC, 'svm'),
}


def get_objective(model_init, scorer, X_train, y_train, X_val, y_val, callbacks=()):
    def objective(params):
        model = model_init(**params)
        model.fit(X_train, y_train)
        preds = model.predict(X_val)
        error = scorer(y_val, preds)
        for callback in callbacks:
            callback(error, params)
        return error

    return objective


def get_scorer(scorer_in):
    if callable(scorer_in):
        return scorer_in
    elif scorer_in == 'neg_rmse':
        return neg_rmse
    else:
        return get_sklearn_scorer(scorer_in)._score_func



def optimize(model_name, scorer, y_column, trials=30, tuner='optuna', feature_selection_support=None,
             feature_selection_cols=None, ds=None, path=None, sample_size=None, stratify=True,
             custom_params={}, ignore=[], n_jobs=-1, logging=None, logging_params={},
             early_stopping=False, patience=100, threshold=1e-3, show_live=True, X_y_train_val=None):
    if X_y_train_val is None:
        X_ds, y_ds = read_dataset(ds, path, y_column, feature_selection_support, feature_selection_cols, ignore,
                                  sample_size, stratify)

        X_train, X_val, y_train, y_val = train_test_split(X_ds, y_ds)
    else:
        X_train, X_val, y_train, y_val = X_y_train_val

    model_init, family = models[model_name]
    custom_params.update(family_params[family])

    callbacks = []
    logger = get_logger(logging, 'feature_analysis', logging_params)
    callbacks.append(lambda error, params: logger(error, params))
    if show_live:
        live = LivePyPlot('maximize')
        callbacks.append(lambda error, params: live(error))
    if early_stopping:
        stopping = EarlyStopping('maximize', patience, threshold)
        callbacks.append(lambda error, params: stopping(error))
    scorer = get_scorer(scorer)

    tuner = get_tuner(tuner, get_objective(model_init, scorer, X_train, y_train, X_val, y_val, callbacks),
                      custom_params)

    try:
        tuner.tune(trials, n_jobs)
    except KeyboardInterrupt:
        logger("Stopped with keyboard")
    except EarlyStoppingError:
        logger("Stopped with early stopping")

    logger(tuner.best_score, tuner.best_params)

    return tuner.best_score, tuner.best_params


def get_tuner(tuner, *params):
    if tuner == 'optuna':
        return OptunaTuner(*params)
    else:
        return RandomSearchTuner(*params)


class Tuner():

    def __init__(self, objective, params_setup):
        self.objective = objective
        self.params_setup = params_setup

    def tune(self, trials, n_jobs=-1):
        pass


class OptunaTuner(Tuner):

    def __init__(self, objective, params_setup, sampler=TPESampler(seed=666)):
        super(OptunaTuner, self).__init__(objective, params_setup)
        self.study = optuna.create_study(direction='maximize', sampler=sampler)
        optuna.logging.set_verbosity(optuna.logging.ERROR)

    def _save_result(self):
        print(self.study.best_params, self.study.best_value)
        self.best_params = self.study.best_params
        self.best_score = self.study.best_value

    def tune(self, trials, n_jobs=-1):
        def optuna_objective(trial):
            params = {col: values[0] if len(values) == 1 else getattr(trial, f'suggest_{values[0]}')(col, *values[1:])
                      for col, values in self.params_setup.items()}
            error = self.objective(params)
            return error

        try:
            self.study.optimize(optuna_objective, n_trials=trials, n_jobs=n_jobs)
        except:
            self._save_result()
            raise
        self._save_result()


class RandomSearchTuner(Tuner):

    def _get_random_attr(self, config):
        type, *params = config
        if type == 'int':
            return np.random.randint(params[0], params[1])
        elif type == 'uniform':
            return np.random.random() * (params[1] - params[0]) + params[0]
        elif type == 'categorical':
            return np.random.choice(params[0])
        return None

    def tune(self, trials, n_jobs=-1):
        self.best_params = {attr: self._get_random_attr(attr_config) for attr, attr_config in self.params_setup.items()}
        self.best_score = -np.inf

        for trial in range(trials):
            params = {attr: self._get_random_attr(attr_config) for attr, attr_config in self.params_setup.items()}
            score = self.objective(params)
            if score > self.best_score:
                self.best_params, self.best_score = params, score



import random
from datetime import datetime
import os
import sqlite3
from sqlite3 import Error


def _gen_id():
    return random.randint(0, 10_000_000)


def get_logger(logger, run_name, params):
    if logger == 'local':
        return LocalLogger(run_name, **params)
    elif logger == 'sqlite':
        return SqliteLogger(run_name, **params)
    else:
        return Logger(run_name, **params)


class Logger:

    def __init__(self, run_name, suppress_std=False, time_format="%m.%d.%Y_%H:%M:%S"):
        self.suppress_std = suppress_std
        self.time_format = time_format
        self.run_name = run_name

    def __call__(self, *args, **kwargs):
        self.log(' '.join([str(x) for x in args]))

    def log(self, message):
        tm = datetime.now().strftime(self.time_format)
        formatted_message = f'{tm}: \t{message}'
        self.std(formatted_message)
        return tm, message, formatted_message

    def std(self, message):
        if not self.suppress_std:
            print(message)

    def end_session(self):
        pass

    def _get_env_info(self):
        timestamp = datetime.now().strftime(self.time_format)
        cwd = os.getcwd()
        dir_path = f'{cwd}/logs' if self.run_name == '' else f'{cwd}/logs/{self.run_name}'
        return timestamp, dir_path


class LocalLogger(Logger):

    def __init__(self, run_name, suppress_std=False):
        super(LocalLogger, self).__init__(run_name, suppress_std)
        timestamp, dir_path = super(LocalLogger, self)._get_env_info()
        os.makedirs(dir_path, exist_ok=True)

        self.file_path = f'{dir_path}/{timestamp}.txt'
        self.file = open(self.file_path, 'a')

        self.std(f'Logs init. See {self.file_path}')

    def log(self, message):
        _, _, formatted_message = super(LocalLogger, self).log(message)
        if not self.file.closed:
            self.file.write(f'{formatted_message}\n')

    def end_session(self):
        if not self.file.closed:
            self.file.close()


class SqliteLogger(Logger):

    CREATE_RUN_TABLE = """
        create table if not exists run (
            id integer PRIMARY KEY, 
            name text,
            tm text not null
        )
        """

    CREATE_LOG_TABLE = """
        create table if not exists log (
            id integer PRIMARY KEY,
            run_id integer not null,
            message text,
            tm text not null,
            FOREIGN KEY (run_id) REFERENCES run (id) 
        )
        """

    ADD_RUN = """
        INSERT INTO run(id, name, tm)
                  VALUES(?,?,?)
        """

    ADD_LOG = """
        INSERT INTO log(id, run_id, message, tm)
                  VALUES(?,?,?,?)
        """

    def __init__(self, run_name, suppress_std=False):
        super(SqliteLogger, self).__init__(run_name, suppress_std)
        timestamp, dir_path = super(SqliteLogger, self)._get_env_info()
        os.makedirs(dir_path, exist_ok=True)
        self._init_db(f'{dir_path}/{run_name}.db')
        self._exec(self.CREATE_RUN_TABLE)
        self._exec(self.CREATE_LOG_TABLE)

        self.run_id = _gen_id()
        self._add_run(self.run_id, run_name, timestamp)

    def log(self, message):
        tm, message, _ = super(SqliteLogger, self).log(message)
        self._add_log(_gen_id(), self.run_id, message, tm)

    def end_session(self):
        if self.conn:
            self.conn.close()

    def _init_db(self, db_path):
        self.db_path = db_path
        self.conn = None
        try:
            self.conn = sqlite3.connect(self.db_path)
            self.std(f'Database init. See {self.db_path}')
        except Error as e:
            self.std(e)

    def _add_run(self, id, name, tm):
        try:
            c = self.conn.cursor()
            c.execute(self.ADD_RUN, (id, name, tm))
            self.conn.commit()
        except Error as e:
            self.std(e)

    def _add_log(self, id, run_id, message, tm):
        try:
            c = self.conn.cursor()
            c.execute(self.ADD_LOG, (id, run_id, message, tm))
            self.conn.commit()
        except Error as e:
            self.std(e)

    def _exec(self, command):
        try:
            c = self.conn.cursor()
            c.execute(command)
            self.conn.commit()
        except Error as e:
            self.std(e)



from sklearn.base import BaseEstimator
import numpy as np
from sklearn.base import clone
from .logs.loggers import get_logger
import math


class DeepModel(BaseEstimator):

    def __init__(self, estimator, depths, n_estimators=100,
                 learning_rate=0.01, verbose=True, logging=None, logging_params={}):
        self.n_estimators = n_estimators
        self.learning_rate = learning_rate
        self.verbose = verbose
        self.depths = depths
        self.estimator = estimator
        self.logger = get_logger(logging, 'DeepModel', logging_params)

    def fit(self, X_train, y_train):
        self.models = []
        self.logger('Training...')
        feed = y_train.copy()
        for depth in self.depths:
            self.logger(f"Depth: {depth}")
            model = clone(self.estimator)
            model.fit(X_train, feed)
            self.models.append(model)
            preds = model.predict(X_train)
            feed -= preds
            self.logger('%.15f' % np.mean(abs(feed)))

    def predict(self, X_test):
        preds = np.zeros(X_test.shape[0])
        for model in self.models:
            preds += model.predict(X_test)
            return preds
        return preds


class EarlyStoppingError(Exception):
    pass


class EarlyStopping:

    def __init__(self, direction, patience=100, threshold=1e-3):
        self.best = -math.inf if direction == 'maximize' else math.inf
        self.fn = max if direction == 'maximize' else min
        self.count = 0
        self.threshold = threshold
        self.patience = patience

    def __call__(self, value):
        new_value = self.fn(self.best, value)
        if abs(new_value - self.best) < self.threshold:
            self.count += 1
            if self.count > self.patience:
                raise EarlyStoppingError()
        else:
            self.count = 0
            self.best = new_value